
from ... classes.flux_row_cls import flux_row_cls


# noinspection PyMissingConstructor
class lev_row_cls(flux_row_cls):

    def __init__(self, headers, values, address=''):
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values
        self.__dict__['address']  = address

    def __repr__(self):
        return '{} {}'.format(self.address, repr(self.values))





import gc

from collections import Counter
from collections import namedtuple
from copy import deepcopy
from types import SimpleNamespace

from .flux_row_cls import flux_row_cls

from ..util.filesystem import parse_file_extension
from ..util.filesystem import read_file
from ..util.filesystem import write_file
from ..util.filesystem import pickle_extensions

from ..util.iter import OrderedDefaultDict
from ..util.iter import IterationDepthError
from ..util.iter import base_class_names
from ..util.iter import is_exhaustable
from ..util.iter import is_collection
from ..util.iter import is_subscriptable
from ..util.iter import is_vengeance_class
from ..util.iter import standardize_variable_arity_arguments
from ..util.iter import iteration_depth
from ..util.iter import map_to_numeric_indices
from ..util.iter import modify_iteration_depth
from ..util.iter import iterator_to_list
from ..util.iter import transpose_as_lists

from ..util.text import print_runtime
from ..util.text import json_dumps_extended
from ..util.text import deprecated
from ..util.text import object_name
from ..util.text import format_vengeance_header

from ..conditional import ordereddict
from ..conditional import line_profiler_installed
from .. conditional import numpy_installed

if numpy_installed:
    import numpy


class flux_cls:
    """ primary data management class

    * converts a list of lists (a matrix) to a list of flux_row_cls objects
    * similar to a pandas DataFrame, but optimized for row-major ordered operations (df.iterrows())

    eg:
        matrix = [['col_a', 'col_b', 'col_c'],
                  ['a', 'b', 'c']]
        flux = flux_cls(matrix)
        flux = flux_cls.from_csv('file.csv')

        for row in flux:
            a = row.col_a
            row.col_a = None
            
        flux['col_d'] = [o.upper() for c in flux['col_a']]
        flux.sort('col_a', 'col_b')
        flux.filter(lambda: row: row.col_a == 'value')
        
        flux.to_csv('file.csv')
        flux = flux_cls.from_csv('file.csv')
    """

    def __init__(self, matrix=None):
        was_gc_enabled = gc.isenabled()
        gc.disable()

        matrix  = self.__validate_matrix(matrix)
        headers = self.__validate_matrix_headers(matrix[0])
        matrix  = [flux_row_cls(headers, row) for row in matrix]

        self.headers = headers
        self.matrix  = matrix

        if was_gc_enabled:
            gc.enable()

    def as_array(self, r_1=1, r_2=None):
        """ to help with debugging: meant to trigger a debugging feature in PyCharm

        PyCharm will recognize the ndarray and enable the "...view as array"
        option in the debugger which displays values in a special window as a table

        deal with jagged rows?
        """
        if not numpy_installed:
            raise ImportError('numpy site-package not installed')

        headers = [format_vengeance_header(n) for n in self.header_names()]
        if self.num_rows <= 0:
            return numpy.array(headers)

        if r_1 < 0:
            r_1 = max(0, self.num_rows + r_1)

        if r_1 == 0:
            r_1 = 1

        headers.insert(0, 'r_i')

        m = [headers]
        for i, row in enumerate(self.matrix[r_1:r_2], r_1):
            values = row.values
            values = ['{:,}'.format(i).replace(',', '_'),
                      *values]
            m.append(values)

        return numpy.array(m, dtype=object)

    def header_names(self):
        """
        self.matrix[0].values and self.headers.keys may not always be identical
            map_numeric_indices() makes certain modifications to self.headers.keys,
            such as coercing values to strings, modifying duplicate values, etc
        """
        return list(self.headers.keys())

    @property
    def num_rows(self):
        """ header not counted """
        if not any(self.matrix[0].values):
            return -1

        return len(self.matrix) - 1

    def __len__(self):
        return self.num_rows

    @property
    def num_cols(self):
        return len(self.matrix[0].values)

    def is_jagged(self):
        num_cols = self.num_cols

        for row in self.matrix[1:]:
            # noinspection PyProtectedMember
            if ( len(row.values)   != num_cols or
                 len(row._headers) != num_cols ):
                return True

        return False

    def is_empty(self):
        for row in self.matrix:
            if any(row.values):
                return False

        return True

    # region {filesystem methods}
    def to_file(self, path,
                      encoding=None,
                      filetype=None,
                      **fkwargs):

        filetype = parse_file_extension((filetype or path), include_dot=True)
        filetype = filetype.lower()

        if filetype == '.csv':
            return self.to_csv(path, encoding, **fkwargs)
        elif filetype == '.json':
            return self.to_json(path, encoding, **fkwargs)
        elif filetype in pickle_extensions:
            return self.serialize(path, **fkwargs)

        raise ValueError("invalid filetype: '{}' \nfiletype must be in {}"
                         .format(filetype, ['.csv', '.json'] + list(pickle_extensions)))

    @classmethod
    def from_file(cls, path,
                       encoding=None,
                       filetype=None,
                       **fkwargs):

        if isinstance(path, flux_cls):
            raise TypeError('function is a classmethod, should be called from flux_cls, not flux')

        filetype = parse_file_extension((filetype or path), include_dot=True)
        filetype = filetype.lower()

        if filetype == '.csv':
            return cls.from_csv(path, encoding, **fkwargs)
        if filetype == '.json':
            return cls.from_json(path, encoding, **fkwargs)
        if filetype in pickle_extensions:
            return cls.deserialize(path, **fkwargs)

        raise ValueError("invalid filetype: '{}' \nfiletype must be in {}"
                         .format(filetype, ['.csv', '.json'] + list(pickle_extensions)))

    def to_csv(self, path,
                     encoding=None,
                     **fkwargs):

        write_file(path, self.rows(), encoding, filetype='.csv', fkwargs=fkwargs)
        return self

    @classmethod
    def from_csv(cls, path,
                      encoding=None,
                      **fkwargs):

        if isinstance(path, flux_cls):
            raise TypeError('function is a classmethod, should be called from flux_cls, not flux')

        m = read_file(path, encoding, filetype='.csv', fkwargs=fkwargs)
        return cls(m)

    def to_json(self, path=None,
                      encoding=None,
                      **fkwargs):

        h = tuple(self.header_names())
        j = [ordereddict(zip(h, row.values)) for row in self.matrix[1:]]

        if path is None:
            return json_dumps_extended(j, **fkwargs)

        write_file(path, j, encoding, filetype='.json', fkwargs=fkwargs)
        return self

    @classmethod
    def from_json(cls, path,
                       encoding=None,
                       **fkwargs):

        if isinstance(path, flux_cls):
            raise TypeError('function is a classmethod, should be called from flux_cls, not flux')

        j = read_file(path, encoding, filetype='.json', fkwargs=fkwargs)

        if not j:
            m = [[]]
        else:
            first_row = j[0]
            if isinstance(first_row, dict):
                m = [list(first_row.keys())] + \
                    [list(row.values()) for row in j]
            else:
                m = j

        return cls(m)

    def serialize(self, path, **fkwargs):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        write_file(path, self, filetype='.flux', fkwargs=fkwargs)
        return self

    @classmethod
    def deserialize(cls, path, **fkwargs):
        """
        although convenient, pickle introduces significant security flaws
        you should be sure no malicious actors have access to the location of these files
        """
        if isinstance(path, flux_cls):
            raise TypeError('function is a classmethod, should be called from flux_cls, not flux')

        return read_file(path, filetype='.flux', fkwargs=fkwargs)
    # endregion

    def columns(self, *names):
        names = standardize_variable_arity_arguments(names, depth=1)

        rva = self.row_values_accessor(names)
        col = (rva(row) for row in self.matrix[1:])

        is_single_column = (iteration_depth(names) == 1
                            and len(names) == 1)
        if not is_single_column:
            col = list(col)
            col = transpose_as_lists(col)

        return col

    def rows(self, r_1=0, r_2=None):
        return (row.values for row in self.matrix[r_1:r_2])

    def flux_rows(self, r_1=0, r_2=None):
        return (row for row in self.matrix[r_1:r_2])

    def namedrows(self, r_1=1, r_2=None):
        h = tuple(self.header_names())
        for row in self.matrix[r_1:r_2]:
            yield SimpleNamespace(**ordereddict(zip(h, row.values)))

    # noinspection PyArgumentList
    def namedtuples(self, r_1=1, r_2=None):
        FluxRow = namedtuple('FluxRow', self.header_names())
        for row in self.matrix[r_1:r_2]:
            yield FluxRow(*row.values)

    def reset_headers(self, names=None):
        """
        simply re-assigning self.headers to a new variable, eg
        self.headers = map_to_numeric_indices(names)), will
        de-reference all flux_row_cls._headers in matrix
        """
        use_existing_headers = (names is None)

        if use_existing_headers:
            names = self.matrix[0].values

        headers = map_to_numeric_indices(names)
        self.headers.clear()
        self.headers.update(headers)

        if not use_existing_headers:
            self.matrix[0].values = list(headers.keys())

        return self

    def reset_matrix(self, m):
        (self.headers,
         self.matrix) = self.__validate_headers_and_matrix(m, headers=None)

        return self

    def execute_commands(self, commands, profiler=False):
        commands = self.__validate_command_methods(commands)
        profiler = self.__validate_profiler_function(profiler)

        completed_commands = []
        for command in commands:
            flux_method = command.method
            if profiler:
                flux_method = profiler(flux_method)

            if command.num_args == 0:
                flux_method()
            elif command.num_args == 1:
                flux_method(command.args)
            else:
                flux_method(*command.args)

            completed_commands.append((command.name, command.args))

        if hasattr(profiler, 'print_stats'):
            profiler.print_stats()

        return completed_commands

    def matrix_by_headers(self, *names):
        if self.is_empty():
            raise ValueError('matrix is empty')

        names = standardize_variable_arity_arguments(names, depth=1)
        if not names:
            return self

        if isinstance(names, dict):
            names = [names]

        headers = self.headers
        names   = [self.__validate_renamed_or_inserted_column(name, headers) for name in names]

        if self.num_rows <= 0:
            (self.headers, self.matrix) = self.__validate_headers_and_matrix([names], headers=None)
            return self

        columns = [row.values for row in self.matrix[1:]]
        columns = list(transpose_as_lists(columns))
        inserted_column = [None for _ in range(self.num_rows)]

        m = []
        for n in names:
            if n in headers:
                column = columns[headers[n]]
            else:
                column = inserted_column

            m.append([n] + column)

        m = list(transpose_as_lists(m))
        (self.headers, self.matrix) = self.__validate_headers_and_matrix(m, headers=None)

        return self

    def rename_columns(self, old_to_new_mapping):
        header_names = self.header_names()

        for h_old, h_new in old_to_new_mapping.items():
            i = self.headers[h_old]
            header_names[i] = h_new

        self.reset_headers(header_names)

        return self

    def insert_columns(self, *names):
        """ eg:
            flux.insert_columns((0, 'inserted'))         insert as first column
            flux.insert_columns((3,  'inserted'))        insert column before 4th column
            flux.insert_columns((-1, 'inserted'))        insert column at end
            flux.insert_columns(('col_c', 'inserted'))   insert column before column 'col_c'

            flux.insert_columns((1, 'inserted_a'),       insert multiple columns before 1st column
                                (1, 'inserted_b'),
                                (1, 'inserted_c'))
        """
        names = standardize_variable_arity_arguments(names, depth=1)
        if not names:
            return self

        names = self.__validate_inserted_items(names, self.headers)
        names = list(reversed(names))

        header_names = self.header_names()
        for before, header in names:
            if isinstance(before, int):
                i = before
            else:
                i = header_names.index(before)

            header_names.insert(i, header)

        indices = sorted([header_names.index(h) for _, h in names])
        self.reset_headers(header_names)

        for i in indices:
            for row in self.matrix[1:]:
                row.values.insert(i, None)

        return self

    def append_columns(self, *names):
        names = standardize_variable_arity_arguments(names, depth=1)
        if not names:
            return self

        self.__validate_no_duplicate_names(names)
        self.__validate_no_names_intersect_with_headers(names, self.headers)

        v = [None for _ in names]
        self.reset_headers(self.header_names() + list(names))

        for row in self.matrix[1:]:
            row.values.extend(v)

        return self

    def delete_columns(self, *names):
        """
        method will fail if columns are jagged, do a try / except on del row.values[i]
        then length check on row values if failure?
        """
        names = standardize_variable_arity_arguments(names, depth=1)
        if not names:
            return self

        indices = self.__validate_names_as_indices(names, self.headers)
        self.__validate_no_duplicate_names(indices)

        all_columns_deleted = (set(self.headers.values()) == set(indices))
        if all_columns_deleted:
            return self.reset_matrix(None)

        indices.sort(reverse=True)
        for i in indices:
            for row in self.matrix:
                del row.values[i]

        self.reset_headers()

        return self

    def jagged_rows(self):
        JaggedRow = namedtuple('JaggedRow', ('i', 'row'))

        num_cols = self.num_cols
        for i, row in enumerate(self.matrix):
            if len(row.values) != num_cols:
                yield JaggedRow(i, row)

    def insert_rows(self, i, rows):

        if self.is_empty():
            (self.headers,
             self.matrix) = self.__validate_headers_and_matrix(rows, headers=None)

            return self

        _, m = self.__validate_headers_and_matrix(rows, self.headers)
        replace_headers = (i == 0)

        if replace_headers:
            self.reset_headers(m.pop(0).values)
            i = 1
        else:
            if len(m) > len(rows):
                del m[0]

            if len(rows) != len(m):
                raise AssertionError('author error')

        self.matrix[i:i] = m

        return self

    def append_rows(self, rows):

        if self.is_empty():
            (self.headers,
             self.matrix) = self.__validate_headers_and_matrix(rows, headers=None)

            return self

        _, m = self.__validate_headers_and_matrix(rows, self.headers)
        if m[0].is_header_row():
            del m[0]

        self.matrix.extend(m)

        return self

    def join(self, other, *on_columns):
        """
        what if user wants another mapping method ?
            d = iterable.map_rows(names_b)
            d = iterable.map_rows_append(names_b)

        or different rowtype?
            d = iterable.map_rows(names_b, rowtype=namedtuple)
        """

        names = modify_iteration_depth(on_columns, depth=0)
        if isinstance(names, dict):
            names_a = list(names.keys())[0]
            names_b = list(names.values())[0]
        else:
            names_a = names
            names_b = names

        self.__validate_all_names_intersect_with_headers(names_a, self.headers)
        if isinstance(other, flux_cls):
            self.__validate_all_names_intersect_with_headers(names_b, other.headers)
            d = other.map_rows(names_b)
        elif isinstance(other, dict):
            d = other
        else:
            d = {item: item for item in other}

        rva = self.row_values_accessor(names_a)

        for row in self.matrix[1:]:
            k = rva(row)
            if k in d:
                yield row, d[k]

    def sort(self, *names, reverse=None):
        """ in-place sort

        eg:
            flux.sort('col_a')
            flux.sort('col_a', 'col_b', 'col_c',
                      reverse=[True, False, True])
        """
        self.matrix[1:] = self.__sort_rows(self.matrix[1:],
                                           names,
                                           reverse)
        return self

    def sorted(self, *names, reverse=None):
        """ :return: sorted flux_cls

        eg:
            flux = flux.sorted('col_a')
            flux = flux.sorted('col_a', 'col_b', 'col_c',
                               reverse=[False, True, True])
        """
        flux = self.copy()
        flux.matrix[1:] = self.__sort_rows(flux.matrix[1:],
                                           names,
                                           reverse)
        return flux

    def __sort_rows(self, rows, names, s_orders):
        """
        names    = reversed(names)
        s_orders = reversed(s_orders)
            # sort priority proceeds in reverse s_orders of names submitted
            # rightmost name sorted first, leftmost name sorted last
        """
        if isinstance(s_orders, bool):
            s_orders = [s_orders]
        else:
            s_orders = s_orders or []

        if not isinstance(s_orders, (list, tuple)):
            raise TypeError('reverse must be a list of booleans')

        names = modify_iteration_depth(names, depth=1)
        for _ in range(len(names) - len(s_orders)):
            s_orders.append(False)

        names    = reversed(names)
        s_orders = reversed(s_orders)

        for n, so in zip(names, s_orders):
            rva = self.row_values_accessor(n)
            so  = bool(so)
            rows.sort(key=rva, reverse=so)

        return rows

    def filter(self, f, *args, **kwargs):
        """ in-place """
        self.matrix[1:] = [row for row in self.matrix[1:]
                               if f(row, *args, **kwargs)]
        return self

    def filtered(self, f, *args, **kwargs):
        """ :return: new flux_cls """
        flux = self.copy()
        flux.matrix[1:] = [row for row in flux.matrix[1:]
                               if f(row, *args, **kwargs)]
        return flux

    def filter_by_unique(self, *names):
        self.__filter_unique_rows(*names, in_place=True)
        return self

    def filtered_by_unique(self, *names):
        return self.__filter_unique_rows(*names, in_place=False)

    def __filter_unique_rows(self, *names, in_place):
        # region {closure functions}
        u   = set()
        rva = self.row_values_accessor(names)

        def evaluate_unique(row):
            v = rva(row)

            if v not in u:
                u.add(v)
                return True
            else:
                return False
        # endregion

        if in_place:
            self.filter(evaluate_unique)
        else:
            return self.filtered(evaluate_unique)

    def map_rows(self, *names, rowtype='flux_row_cls'):
        """ :return: dictionary of {row_value(row): row} """
        items = self.__map_row_items(names, rowtype=rowtype)
        return ordereddict(items)

    def map_rows_append(self, *names, rowtype='flux_row_cls'):
        """ :return: dictionary of {row_value(row): [rows]} """
        items = self.__map_row_items(names, rowtype=rowtype)

        return (OrderedDefaultDict(default=list)
                .append_items(items)
                .ordereddict())

    @deprecated('Use flux_cls.map_rows() method instead')
    def index_row(self, *names, rowtype='flux_row_cls'):
        """ deprecated """
        return self.map_rows(*names, rowtype=rowtype)

    @deprecated('Use flux_cls.map_rows_append() method instead')
    def index_rows(self, *names, rowtype='flux_row_cls'):
        """ deprecated """
        return self.map_rows_append(*names, rowtype=rowtype)

    def __map_row_items(self, names, rowtype):
        rowtype = self.__validate_mapped_rowtype(rowtype)

        rva  = self.row_values_accessor(names)
        rows = iter(self.matrix[1:])

        if rowtype == 'flux_row_cls':
            items = ((rva(row), row) for row in rows)
        elif rowtype == 'namedrow':
            items  = ((rva(row), v) for row, v in zip(rows, self.namedrows()))
        elif rowtype == 'namedtuple':
            items  = ((rva(row), v) for row, v in zip(rows, self.namedtuples()))
        elif rowtype == 'list':
            items = ((rva(row), row.values) for row in rows)
        elif rowtype == 'tuple':
            items = ((rva(row), tuple(row.values)) for row in rows)
        else:
            raise AssertionError

        return items

    def unique(self, *names):
        rva   = self.row_values_accessor(names)
        items = ((rva(row), None) for row in self.matrix[1:])

        return ordereddict(items).keys()

    def contiguous(self, *names):
        """ :return: yield (value, i_1, i_2) namedtuple where values are contiguous

        (contiguous values may only span a single row ie, i_1 == i_2)
        """
        if self.num_rows <= 0:
            raise IndexError('matrix has no rows')

        ContiguousRows = namedtuple('ContiguousRows', ('i_1', 'i_2', 'rows'))

        i_1 = 1

        rva = self.row_values_accessor(names)
        v_1 = rva(self.matrix[1])

        for i_2, row in enumerate(self.matrix[2:], 2):
            v_2 = rva(row)

            if v_2 != v_1:
                yield ContiguousRows(i_1, i_2 - 1, self.matrix[i_1:i_2])
                v_1 = v_2
                i_1 = i_2

        yield ContiguousRows(i_1, self.num_rows, self.matrix[i_1:])

    def label_row_indices(self, start=0):
        """ meant to assist with debugging;

        label each flux_row_cls.__dict__ with an index, which will then appear
        in each row's __repr__ function and make them easier to identify after
        filtering, sorting, etc
        """
        if 'r_i' in self.headers:
            raise NameError("'r_i' already exists as header name")

        for i, row in enumerate(self.matrix, start):
            row.__dict__['r_i'] = i

        return self

    def copy(self, deep=False):
        if deep:
            return deepcopy(self)

        m = [[*row.values] for row in self.matrix]
        flux = self.__class__(m)

        return flux

    def row_values_accessor(self, names):
        """ :return: a function that retrieves column values for each row in self.matrix """

        # region {closure functions}
        def row_value(row):
            """ return single column value from each row """
            return row.values[i]

        def row_values(row):
            """ return muliple column values from each row """
            return tuple([row.values[_i_] for _i_ in i])

        def row_values_slice(row):
            return tuple(row.values[i])

        def row_values_all(row):
            """ return muliple column values from each row """
            return tuple(row.values)
        # endregion

        if is_exhaustable(names):
            names = list(names)

        if were_names_submitted_as_tuples(names):
            names = modify_iteration_depth(names, depth=1)
        else:
            names = iterator_to_list(names)
            names = modify_iteration_depth(names, depth=0)
            names = iterator_to_list(names)

        i = None

        if callable(names):
            f = names
        elif isinstance(names, slice):
            i = names
            f = row_values_slice
        elif iteration_depth(names) == 0:
            i = self.__validate_names_as_indices([names], self.headers)[0]
            f = row_value
        else:
            i = self.__validate_names_as_indices(names, self.headers)
            f = row_values

            if are_indices_contiguous(i):
                i = slice(i[0], i[-1]+1)
                f = row_values_slice

        if isinstance(i, slice):
            if i.start == 0 and i.stop == self.num_cols:
                f = row_values_all

        return f

    def __getitem__(self, name):
        """
        MUST return a flat list if name is not a slice
            flux['col_a'] = flux['col_b']
        """
        if isinstance(name, slice):
            name = self.header_names()[name]

        return self.columns(name)

    def __setitem__(self, name, values):
        """ sets values to a single column
        values are expected to exclude header row

        eg:
            flux['col'] = ['blah'] * flux.num_rows
            flux[-1]    = ['blah'] * flux.num_rows

            to insert column:
                flux[(0, 'new_col')] = ['blah'] * flux.num_rows
        """
        if isinstance(values, str):
            raise IterationDepthError('column values need to be an iterable')

        if isinstance(name, tuple) and len(name) == 2:
            self.insert_columns(name)
            name = name[-1]
        elif isinstance(name, slice):
            raise ValueError('__setitem__ cannot be used to set values to multiple columns')
        elif iteration_depth(name) > 0:
            raise ValueError('__setitem__ cannot be used to set values to multiple columns')

        if name not in self.headers and not isinstance(name, int):
            self.append_columns(name)

        i = self.__validate_names_as_indices([name], self.headers)[0]

        for row, v in zip(self.matrix[1:], values):
            row.values[i] = v

    def __bool__(self):
        return self.is_empty()

    def __iter__(self):
        return (row for row in self.matrix[1:])

    def __add__(self, rows):
        flux = self.copy()
        flux.append_rows(rows)

        return flux

    def __iadd__(self, rows):
        self.append_rows(rows)

        return self

    def __getstate__(self):
        """ serialize raw matrix values, not self.__dict__
        this has advantages in speed and stability
        """
        return list(self.rows())

    def __setstate__(self, o):
        if isinstance(o, dict):
            self.__dict__ = o                   # copy attributes
        else:
            self.__init__(o)                    # call flux_cls constructor

    def __repr__(self):
        if self.is_empty():
            return format_vengeance_header(' ')

        if self.is_jagged():
            is_jagged = '🗲jagged🗲  '
        else:
            is_jagged = ''

        num_rows = '⟨{:,}⟩  '.format(self.num_rows).replace(',', '_')

        headers = ', '.join(str(n) for n in self.header_names())
        headers = format_vengeance_header(headers)

        return '{}{}{}'.format(is_jagged, num_rows, headers)

    # region {validation functions}
    @staticmethod
    def __validate_matrix_headers(names):
        names = map_to_numeric_indices(names)

        conflicting = names.keys() & set(flux_row_cls.reserved_names())
        if conflicting:
            raise ValueError('column name(s) conflict with existing headers: {}'.format(conflicting))

        return names

    @staticmethod
    def __validate_matrix(m):
        if (m is None) or (m == []):
            return [[]]
        if is_vengeance_class(m):
            return [[*row] for row in m.rows()]

        if is_exhaustable(m):
            m = list(m)

        base_names = set(base_class_names(m))
        if 'DataFrame' in base_names:
            raise NotImplementedError('matrix as DataFrame not supported')
        elif 'ndarray' in base_names:
            raise NotImplementedError('matrix as ndarray not supported')

        if not is_subscriptable(m):
            raise IndexError("invalid type: <'{}'> matrix expected to be a list of lists".format(type(m)))

        first_row = m[0]

        # list of flux_row_cls
        if isinstance(first_row, flux_row_cls):
            m = [[*row.values] for row in m]

            if not first_row.is_header_row():
                m.insert(0, first_row.header_names())

        # list of namedtuples
        elif hasattr(first_row, '_fields'):
            # noinspection PyProtectedMember
            m = [list(first_row._fields)] + \
                [list(row) for row in m]

        # list of objects
        elif hasattr(first_row, '__dict__'):
            m = [list(first_row.__dict__.keys())] + \
                [list(row.__dict__.values()) for row in m]

        # json or column-major?
        # elif isinstance(first_row, dict):
        #     m = [list(first_row.keys())] + \
        #         [list(row.values()) for row in m]

        elif iteration_depth(m, first_element_only=True) < 2:
            raise IterationDepthError('matrix must have at least two dimensions (ie, a list of lists)')

        return m

    @staticmethod
    def __validate_headers_and_matrix(matrix, headers):
        matrix = flux_cls.__validate_matrix(matrix)
        if headers is None:
            headers = flux_cls.__validate_matrix_headers(matrix[0])

        return headers, [flux_row_cls(headers, row) for row in matrix]

    def __validate_command_methods(self, commands):
        Command = namedtuple('Command', 'name method args num_args')

        parsed = []
        for command in commands:

            if isinstance(command, (list, tuple)):
                name, args = command
                if isinstance(args, dict):
                    raise NotImplementedError('keyword args not supported')

                num_args = iteration_depth(args) + 1
            else:
                name = command
                args = ()
                num_args = 0

            if name.startswith('__'):
                name = '_{}{}'.format(self.__class__.__name__, name)

            method = getattr(self, name)
            parsed.append(Command(name, method, args, num_args))

        return parsed

    @staticmethod
    def __validate_profiler_function(use_profiler):
        if use_profiler in (None, False):
            return None

        if use_profiler is True:
            if line_profiler_installed:
                from line_profiler import LineProfiler
                return LineProfiler()
            else:
                return print_runtime

        profiler = str(use_profiler).lower()

        if profiler == 'print_runtime':
            return print_runtime

        if profiler in ('line_profiler', 'lineprofiler'):
            if line_profiler_installed is False:
                raise ImportError("'line_profiler' package not installed")

            from line_profiler import LineProfiler
            return LineProfiler()

        raise ValueError("invalid profiler: '{}', profiler should be in "
                         "\n(None, False, True, 'print_runtime', 'line_profiler')".format(profiler))

    @staticmethod
    def __validate_renamed_or_inserted_column(name, headers):
        is_remapped_column = isinstance(name, dict)
        is_inserted_column = (isinstance(name, str)   and name.startswith('(')  and name.endswith(')') or
                              isinstance(name, bytes) and name.startswith(b'(') and name.endswith(b')')
                              and name not in headers)

        if is_remapped_column:
            name_old, name = list(name.items())[0]
            headers[name] = headers[name_old]
        elif is_inserted_column:
            name = name[1:-1]
            if name in headers:
                raise ValueError("column: '{}' already exists".format(name))
        elif name not in headers:
            raise KeyError("column: '{name}' does not exist. "
                           "\nTo ensure new columns are being created intentionally (not a new column "
                           "because to a typo) inserted headers must be surrounded by parenthesis, eg: "
                           "\n '({name})', not '{name}'".format(name=name))

        return name

    @staticmethod
    def __validate_mapped_rowtype(rowtype):
        validtypes = {'flux_row_cls',
                      'namedrow',
                      'namedtuple',
                      'tuple',
                      'list'}

        if not isinstance(rowtype, str):
            rowtype = object_name(rowtype)

        if rowtype.lower() not in validtypes:
            raise TypeError('rowtype parameter must be in the following: \n'
                            '{}'.format('\n\t'.join(validtypes)))

        return rowtype

    @staticmethod
    def __validate_name_datatypes(names, headers=None):
        valid_datatypes = (int, 
                           str, 
                           bytes)
        
        names = standardize_variable_arity_arguments(names, depth=1)

        if headers is not None:
            flux_cls.__validate_names_not_empty(names)

            if isinstance(names[0], slice):
                names = names[0]
                names = list(headers.values())[names]
        
        invalid = [n for n in names if not isinstance(n, valid_datatypes)]
        if invalid:
            s = ("invalid name datatype: '{}' "
                 "\nvalid datatypes are {}"
                 '\n'
                 '\n\t(make sure any arguments following variable position parameters are submitted by keyword)'
                 "\n\t\tflux.sort('column_a', 'column_b', reverse=True) "
                 '\n\tand not: '
                 "\n\t\tflux.sort('column_a', 'column_b', True)"
                 "\n".format(invalid, valid_datatypes))
            
            raise TypeError(s)

        return names

    @staticmethod
    def __validate_names_not_empty(names):
        if len(names) == 0:
            raise ValueError('no column names submitted')

    @staticmethod
    def __validate_inserted_items(inserted, headers):
        inserted = standardize_variable_arity_arguments(inserted, depth=2)

        for item in inserted:
            if iteration_depth(item) != 1 or len(item) != 2:
                raise IterationDepthError("inserted values must be ({location}, {name}) tuples \n\teg: (2, 'new_col')")

        indices, names = list(zip(*inserted))

        flux_cls.__validate_no_duplicate_names(names)
        flux_cls.__validate_no_names_intersect_with_headers(names, headers)
        flux_cls.__validate_all_names_intersect_with_headers(indices, headers)

        return inserted

    @staticmethod
    def __validate_no_duplicate_names(names):
        names = flux_cls.__validate_name_datatypes(names)

        duplicates = [n for n, count in Counter(names).items() if count > 1]
        if duplicates:
            raise ValueError('duplicate column name(s) detected: \n{}'.format(duplicates))

        return names

    @staticmethod
    def __validate_no_names_intersect_with_headers(names, headers):
        names = flux_cls.__validate_name_datatypes(names, headers)

        conflicting = headers.keys() & set(names)
        if conflicting:
            raise ValueError('column name(s) conflict with existing headers: \n{}'.format(conflicting))

        return names

    @staticmethod
    def __validate_all_names_intersect_with_headers(names, headers):
        names = flux_cls.__validate_name_datatypes(names, headers)

        h_names = [*[ i for i in range(len(headers))],
                   *[-i for i in range(len(headers) + 1)]]
        h_names = headers.keys() | set(h_names)

        invalid = set(names) - h_names
        if invalid:
            s = '\n\t'.join(str(n) for n in headers.keys())
            s = ("\ncolumn name(s) not found: '{}' "
                 "\navailable columns: "
                 "\n\t{}".format(invalid, s))

            raise ValueError(s)

        return names

    @staticmethod
    def __validate_names_as_indices(names, headers):
        names = flux_cls.__validate_all_names_intersect_with_headers(names, headers)

        h_indices = list(headers.values())
        indices = [headers.get(n, n) for n in names]       # lookup strings
        indices = [h_indices[i] for i in indices]          # lookup negative integers

        return indices
    # endregion


def were_names_submitted_as_tuples(n):
    """
    if names were submitted as tuples, row values should be returned as tuples,
    regardless if names contain only a single column

    names submitted as tuples:
        flux.map_rows(('col_a',))
    names not submitted as tuples:
        flux.map_rows('col_a')

    eg:
        True  = were_names_submitted_as_tuples( (('col_a',),) )
        False = were_names_submitted_as_tuples( ('col_a',) )
    """
    if not is_collection(n):
        return False

    return iteration_depth(n) >= 2 and isinstance(n[0], tuple)


def are_indices_contiguous(indices):
    differences = [i_2 - i_1 for i_2, i_1 in zip(indices[1:], indices)]
    return set(differences) == {1}





import csv
import gc
import re
import os
import pprint
import shutil

from collections import namedtuple
from datetime import datetime
from glob import glob

from . text import json_unhandled_conversion
from .. conditional import cpickle
from .. conditional import ultrajson_installed

if ultrajson_installed:
    import ujson as json
else:
    import json

pickle_extensions = {'.flux', '.pkl', '.pickle'}


# noinspection DuplicatedCode
def read_file(path,
              encoding=None,
              mode='r',
              filetype=None,
              **fkwargs):

    if mode[0] != 'r' or mode[-1] == '+':
        raise ValueError('invalid read mode: {}'.format(mode))

    filetype = __validate_filetype(filetype, path)
    as_bytes = __validate_mode_with_encoding(mode, encoding, filetype)
    mode     = __validate_mode_with_filetype(as_bytes, mode, filetype)
    fkwargs  = fkwargs or {}

    if 'fkwargs' in fkwargs:
        fkwargs = fkwargs['fkwargs']

    was_gc_enabled = gc.isenabled()
    gc.disable()

    if filetype == '.csv':
        fkwargs['strict'] = fkwargs.get('strict', True)
        fkwargs['nrows']  = fkwargs.get('nrows')
        nrows = fkwargs.pop('nrows')

        with open(path, mode, encoding=encoding) as f:
            data = csv.reader(f, **fkwargs)

            if nrows is None:
                data = list(data)
            else:
                data = list(__read_limited_csv_rows(data, nrows))

    elif filetype == '.json':
        with open(path, mode, encoding=encoding) as f:
            data = json.load(f, **fkwargs)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            data = cpickle.load(f, **fkwargs)

    elif as_bytes:
        with open(path, mode) as f:
            data = f.read()

    else:
        with open(path, mode, encoding=encoding) as f:
            data = f.read()

    if was_gc_enabled:
        gc.enable()

    return data


# noinspection DuplicatedCode
def write_file(path,
               data,
               encoding=None,
               mode='w',
               filetype=None,
               **fkwargs):

    if not mode[0] in ('w', 'a'):
        raise ValueError('invalid write mode: {}'.format(mode))

    filetype = __validate_filetype(filetype, path)
    as_bytes = __validate_mode_with_encoding(mode, encoding, filetype)
    mode     = __validate_mode_with_filetype(as_bytes, mode, filetype)
    fkwargs  = fkwargs or {}

    if 'fkwargs' in fkwargs:
        fkwargs = fkwargs['fkwargs']

    was_gc_enabled = gc.isenabled()
    gc.disable()

    if filetype == '.csv':
        fkwargs['strict']         = fkwargs.get('strict', True)
        fkwargs['lineterminator'] = fkwargs.get('lineterminator', '\n')

        with open(path, mode, encoding=encoding) as f:
            csv.writer(f, **fkwargs).writerows(data)

    elif filetype == '.json':
        fkwargs['ensure_ascii'] = fkwargs.get('ensure_ascii', encoding in (None, 'ascii'))
        fkwargs['indent']       = fkwargs.get('indent', 4)
        fkwargs['default']      = fkwargs.get('default', json_unhandled_conversion)

        if ultrajson_installed:
            del fkwargs['default']

        with open(path, mode, encoding=encoding) as f:
            json.dump(data, f, **fkwargs)

    elif filetype in pickle_extensions:
        with open(path, mode) as f:
            cpickle.dump(data, f, **fkwargs)

    elif as_bytes:
        with open(path, mode) as f:
            f.write(data)

    else:
        if not isinstance(data, (str, bytes)):
            # f.write() is much faster than f.writelines()
            data = pprint.pformat(data, **fkwargs)
            data = data.replace("'", '"')

        with open(path, mode, encoding=encoding) as f:
            f.write(data)

    if was_gc_enabled:
        gc.enable()


def __validate_filetype(filetype, path):
    """ check for file types that require specialized io libraries / protocols """
    notimplemented_extensions = {'.7z',
                                 '.gzip',
                                 '.png',
                                 '.jpg',
                                 '.jpeg',
                                 '.gif',
                                 '.pdf'}

    filetype = filetype or parse_file_extension(path, include_dot=True)
    filetype = filetype.lower().strip()

    if not filetype.startswith('.'):
        filetype = '.' + filetype

    if filetype.startswith('.xl') or filetype in notimplemented_extensions:
        raise NotImplementedError("file type not supported: '{}'".format(filetype))

    return filetype


def __validate_mode_with_encoding(mode, encoding, filetype):
    as_bytes = mode.replace('+', '').endswith('b')

    if as_bytes and encoding:
        raise ValueError('as bytes mode does not accept an encoding argument')

    if as_bytes and filetype == '.csv':
        raise ValueError('as bytes mode is incompatable with csv module')

    return as_bytes


def __validate_mode_with_filetype(as_bytes, mode, filetype):
    if not as_bytes and (filetype in pickle_extensions):
        if mode.endswith('+'):
            mode = mode[0] + 'b+'
        else:
            mode += 'b'

    return mode


def __read_limited_csv_rows(csv_reader, nrows):
    try:
        for _ in range(nrows):
            yield next(csv_reader)
    except StopIteration:
        pass


def clear_dir(filedir):
    filedir = standardize_dir(filedir, explicit_cwd=True)

    if not os.path.exists(filedir):
        return

    if not os.path.isdir(filedir):
        raise TypeError('"{}" is not a directory'.format(filedir))

    for item in os.listdir(filedir):
        path = filedir + item

        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.remove(path)


def copy_dir(s_dir, d_dir, ignore=None, dirs_exist_ok=False):
    """
    shutil.copytree will fail if shutil.rmtree hasn't finished deleting folder:
        PermissionError: [WinError 5] Access is denied
    """
    s_dir = standardize_dir(s_dir)
    d_dir = standardize_dir(d_dir)

    if ignore:
        if isinstance(ignore, str):
            ignore = (ignore,)
        elif isinstance(ignore, set):
            ignore = tuple(ignore)

        ignore = shutil.ignore_patterns(*ignore)

    if not dirs_exist_ok and os.path.exists(d_dir):
        shutil.rmtree(d_dir)

    shutil.copytree(src=s_dir,
                    dst=d_dir,
                    ignore=ignore,
                    dirs_exist_ok=dirs_exist_ok)


def file_creation_datetime(path):
    unix_t = os.path.getctime(path)
    return datetime.fromtimestamp(unix_t)


def file_modified_datetime(path):
    unix_t = os.stat(path).st_mtime
    return datetime.fromtimestamp(unix_t)


def parse_path(path,
               pathsep='/',
               explicit_cwd=False):

    ParsedPath = namedtuple('ParsedPath', ('directory',
                                           'filename',
                                           'extension'))
    if os.path.isdir(path):
        filedir   = path
        filename  = ''
        extension = ''
    else:
        filedir,  filename  = os.path.split(path)
        filename, extension = os.path.splitext(filename)

    filedir = standardize_dir(filedir, pathsep, explicit_cwd)

    return ParsedPath(filedir, filename, extension)


def parse_file_extension(filename, include_dot=True):
    _, extension = os.path.splitext(filename)
    if (include_dot is False) and extension.startswith('.'):
        extension = extension[1:]

    return extension


def standardize_path(path,
                     pathsep='/',
                     explicit_cwd=False):

    p_path = parse_path(path, pathsep, explicit_cwd)
    path   = ''.join(p_path)

    return path


def standardize_dir(filedir,
                    pathsep='/',
                    explicit_cwd=False):

    if (explicit_cwd is False) and (filedir == ''):
        return filedir

    filedir = filedir or os.getcwd()
    filedir = (filedir.replace('\\', pathsep)
                      .replace('/', pathsep))

    if not filedir.endswith(pathsep):
        filedir += pathsep

    if re.search(r'^[a-z][:][/\\]', filedir):
        filedir = filedir[0].upper() + filedir[1:]

    return filedir


def validate_path_exists(path):

    (filedir,
     filename,
     extension) = parse_path(path, explicit_cwd=True)

    if not os.path.exists(filedir):
        raise FileNotFoundError('directory not found: \n\t{}'.format(filedir))

    path = filedir + filename + extension

    if not os.path.exists(path):
        glob_paths = glob(filedir + filename + '.*')
        if glob_paths:

            extension = parse_file_extension(os.path.split(glob_paths[0])[1], include_dot=True)
            raise FileNotFoundError('file extension not found: '
                                    '\n\t{}'
                                    '\n\t{}'
                                    '\n\tdid you mean: {}?'.format(filedir, filename, extension))

        raise FileNotFoundError('file not found: '
                                '\n\t{}'
                                '\n\t{}'.format(filedir, filename + extension))

    return path








from copy import copy
from collections import namedtuple
from types import SimpleNamespace

from ..util.iter import map_to_numeric_indices
from ..util.text import format_vengeance_header

from ..conditional import ordereddict
from ..conditional import numpy_installed

if numpy_installed:
    import numpy


class flux_row_cls:
    @classmethod
    def reserved_names(cls):
        reserved = [v for v in vars(flux_row_cls).keys()
                      if not v.startswith('_flux_row_cls')]
        reserved.append('_headers')
        reserved.append('values')

        reserved.sort()

        return reserved

    def __init__(self, headers, values):
        """
        :param headers: OrderedDict of {'header': int}
            headers is a single dictionary passed byref from the flux_cls to many flux_row_cls instances
            this eliminates need for all flux_row_cls objects to maintain a seprate copy of these mappings and
            allows for centralized and instantaneous updatdes
        :param values: list of underlying data

        properties must be set on self.__dict__ instead of directly on self to prevent
        premature __setattr__ lookups
        """
        self.__dict__['_headers'] = headers
        self.__dict__['values']   = values

    @property
    def as_array(self):
        """ to help with debugging: meant to trigger a debugging feature in PyCharm

        PyCharm will recognize the ndarray and enable the "...view as array"
        option in the debugger which displays values in a special window as a table
        """
        if not numpy_installed:
            raise ImportError('numpy site-package not installed')

        names   = [format_vengeance_header(n) for n in self.header_names()]
        values  = list(self.values)
        max_len = max(len(names), len(values))

        names.extend(['🗲'] * (max_len - len(names)))
        values.extend(['🗲jagged🗲'] * (max_len - len(values)))

        names  = numpy.array(names,  dtype=object)
        values = numpy.array(values, dtype=object)

        return numpy.transpose([names, values])

    @property
    def headers(self):
        return copy(self._headers)

    def header_names(self):
        return list(self._headers.keys())

    def is_jagged(self):
        return len(self._headers) != len(self.values)

    def is_empty(self):
        return (len(self._headers) == 0) and (len(self.values) == 0)

    def is_header_row(self):
        """ determine if underlying values match self._headers.keys

        self.names == self.values will not always work, since map_numeric_indices()
        was used to modify self._headers values into more suitable dictionary keys,
        like modifying duplicate values to ensure they are unique, etc
        """
        if not self._headers:
            return False

        names = map_to_numeric_indices(self.values)
        return self._headers.keys() == names.keys()

    def dict(self):
        names = self.header_names()
        return ordereddict(zip(names, self.values))

    def namedrow(self):
        return SimpleNamespace(**self.dict())

    # noinspection PyArgumentList
    def namedtuple(self):
        FluxRow = namedtuple('FluxRow', self.header_names())
        return FluxRow(*self.values)

    # noinspection PyProtectedMember,PyUnusedLocal
    def join_values(self, row_b, on_columns=None):
        """ copies all values where header names are shared with row_b
        :type row_b: flux_row_cls
        :type on_columns: shared column names
        """
        # region {closure function}
        def intersecting_names(_names_):
            if not _names_:
                _names_ = headers_a.keys() & headers_b.keys()
                if not _names_:
                    raise ValueError('no intersecting column names')

            return _names_
        # endregion

        if not isinstance(row_b, flux_row_cls):
            raise TypeError('row expected to be flux_row_cls')

        headers_a = self._headers
        headers_b = row_b._headers

        values_a  = self.values
        values_b  = row_b.values

        names = intersecting_names(on_columns)

        for name in names:
            i_b = headers_b[name]
            i_a = headers_a[name]

            values_a[i_a] = values_b[i_b]

    def __getattr__(self, name):
        """  eg:
             o = row.column
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError):
            raise AttributeError(self.__name_not_exist_message(name, self.header_names())) from None

    def __getitem__(self, name):
        """ eg:
            o = row['column']
        """
        try:
            i = self._headers.get(name, name)
            return self.values[i]
        except (TypeError, IndexError):
            if isinstance(name, slice):
                return self.values[name]

            raise AttributeError(self.__name_not_exist_message(name, self.header_names())) from None

    def __setattr__(self, name, value):
        """ eg:
            row.column = o
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError):
            if name in self.__dict__:
                self.__dict__[name] = value
            else:
                raise AttributeError(self.__name_not_exist_message(name, self.header_names())) from None

    def __setitem__(self, name, value):
        """ eg:
            row['column'] = o
        """
        try:
            i = self._headers.get(name, name)
            self.values[i] = value
        except (TypeError, IndexError):
            if name in self.__dict__:
                self.__dict__[name] = value
            elif isinstance(name, slice):
                self.values[name] = value
            else:
                raise AttributeError(self.__name_not_exist_message(name, self.header_names())) from None

    def __len__(self):
        return len(self.values)

    def __bool__(self):
        return bool(self.values)

    def __iter__(self):
        return iter(self.values)

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return id(self._headers) + hash(tuple(self.values))

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, d):
        self.__dict__ = d

    def __repr__(self):
        if 'r_i' in self.__dict__:
            row_index = 'r_i: {:,}    '.format(self.__dict__['r_i']).replace(',', '_')
        else:
            row_index = ''

        if self.is_jagged():
            is_jagged = '🗲jagged🗲  '
        else:
            is_jagged = ''

        if self.is_header_row():
            values = ', '.join(str(n) for n in self.header_names())
            values = format_vengeance_header(values)
        else:
            values = (repr(self.values).replace('"', '')
                                       .replace("'", ''))

        return '{}{}{}'.format(row_index, is_jagged, values)

    @staticmethod
    def __name_not_exist_message(invalid_name, header_names):
        if isinstance(invalid_name, slice):
            return 'slice should be used directly on row.values\n(eg, row.values[2:5], not row[2:5])'

        s = '\n\t'.join(str(n) for n in header_names)
        s = ("\ncolumn name not found: '{}' "
             "\navailable columns: "
             "\n\t{}".format(invalid_name, s))

        return s







from collections import namedtuple
from datetime import date
from datetime import datetime
from datetime import timedelta
from functools import lru_cache

from ..conditional import dateutil_installed

if dateutil_installed:
    from dateutil.parser import parse as dateutil_parse

# 2**15: ~32,800 days, ~90 years of dates
dates_cachesize = 2**15
excel_epoch     = datetime(1900, 1, 1)


def to_datetime(v, d_format=None):
    """
    :param v:        value to be converted
    :param d_format: datetime.strptime format
    """
    if isinstance(v, str):
        date_time = parse_date_string(v, d_format)
    elif isinstance(v, (int, float)):
        date_time = (parse_date_timestamp(v) or
                     parse_date_excel_serial(v) or
                     parse_date_numeric_string(v))

    elif isinstance(v, (list, tuple)):
        date_time = [to_datetime(d, d_format) for d in v]
    elif isinstance(v, datetime):
        date_time = v
    elif type(v) == date:
        date_time = datetime(v.year, v.month, v.day)
    else:
        raise TypeError("can't convert '{}' instance to datetime".format(type(v)))

    if date_time is None:
        raise ValueError("can't convert value: '{}' to datetime".format(v))

    return date_time


def parse_timedelta(td):
    if not isinstance(td, timedelta):
        raise TypeError('value is not timedelta instance')

    ParsedTimeDelta = namedtuple('ParsedTimeDelta', ('days',
                                                     'hours',
                                                     'minutes',
                                                     'seconds'))
    s = td.total_seconds()
    m, s = divmod(s, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    return ParsedTimeDelta(d, h, m, s)


@lru_cache(maxsize=dates_cachesize)
def is_date(v):
    """ :return: (bool success, converted value) """
    try:
        return True, to_datetime(v)
    except (ValueError, TypeError):
        return False, v


@lru_cache(maxsize=dates_cachesize)
def parse_date_timestamp(v):
    """ eg:
        datetime.datetime(2000, 1, 1, 0, 0) = parse_date_timestamp(946702800.0)
    """
    try:
        return datetime.fromtimestamp(v)
    except ValueError:
        return None


@lru_cache(maxsize=dates_cachesize)
def parse_date_excel_serial(v):
    """ number of days since 1900-01-01 """
    try:
        return excel_epoch + timedelta(days=v)
    except ValueError:
        return None


def to_excel_serial(v):
    v = to_datetime(v)
    return (v - excel_epoch).days


@lru_cache(maxsize=dates_cachesize)
def parse_date_numeric_string(v):
    """ eg:
        datetime.datetime(2000, 12, 1, 0, 0) = parse_date_numeric(20001201)
    """
    s = str(int(v))
    if len(s) != 8:
        return None

    try:
        y, m, d = int(s[:4]), int(s[4:6]), int(s[6:8])
        return datetime(y, m, d)
    except ValueError:
        return None


@lru_cache(maxsize=dates_cachesize)
def parse_date_string(s, d_format):
    if d_format is not None:
        return datetime.strptime(s, d_format)

    date_time = __parse_date_strptime(s)

    if date_time is None:
        if dateutil_installed:
            date_time = __parse_date_dateutil(s)
        else:
            from .text import vengeance_message
            print(vengeance_message("('python-dateutil' package not installed)"))

    return date_time


def __parse_date_strptime(s):
    common_formats = ('%m-%d-%Y',       # 01-01-2000
                      '%m/%d/%Y',       # 01/01/2000
                      '%Y-%m-%d',       # 2000-01-01
                      '%Y/%m/%d',       # 2000/01/01
                      '%Y-%b-%d',       # 2000-Jan-01
                      '%d-%b-%Y',       # 01-Jan-2000
                      '%Y%m%d')         # 20000101

    if 'T' in s:
        try:
            return datetime.fromisoformat(s)
        except ValueError:
            return None

    if not (__compatible_with_strptime_lengths(s, common_formats)):
        return None

    for d_format in common_formats:
        try:
            return datetime.strptime(s, d_format)
        except ValueError:
            pass

    return None


def __compatible_with_strptime_lengths(s, common_formats):
    _common_formats_ = [df.replace('%Y', '2000')
                          .replace('%m', '01')
                          .replace('%d', '01')
                          .replace('%b', 'jan') for df in common_formats]

    max_cf = max(len(cf) for cf in _common_formats_)
    min_cf = min(len(cf) for cf in _common_formats_) - 2
    len_s = len(s)

    return min_cf <= len_s <= max_cf


def __parse_date_dateutil(s):
    try:
        return dateutil_parse(s)
    except ValueError:
        return None





from .dates import is_date
from .dates import to_datetime

from .filesystem import read_file
from .filesystem import write_file

from .iter import OrderedDefaultDict
from .iter import transpose

from .text import print_runtime
from .text import print_performance
from .text import styled


__all__ = ['print_runtime',
           'print_performance',
           'styled',
           'to_datetime',
           'is_date',
           'read_file',
           'write_file',
           'transpose',
           'OrderedDefaultDict']




import os
import sys
from collections import OrderedDict

python_version      = sys.version_info
is_pypy_interpreter = ('__pypy__' in sys.builtin_module_names)
is_windows_os       = (os.name == 'nt')
is_tty_console      = (hasattr(sys.stdout, 'isatty') and sys.stdout.isatty())
is_utf_console      = ('utf' in sys.stdout.encoding.lower())

''' ordereddict:
    starting at python 3.6 the built-in dict is both 
    insertion-ordered AND compact, using about half the 
    memory of collections.OrderedDict
'''
if python_version >= (3, 6):
    ordereddict = dict
else:
    ordereddict = OrderedDict

try:
    import _pickle as cpickle
except ImportError:
    import pickle as cpickle

try:
    import dateutil
    dateutil_installed = True
except ImportError:
    dateutil_installed = False

try:
    import ujson
    ultrajson_installed = True
except ImportError:
    ultrajson_installed = False

try:
    import numpy
    numpy_installed = True
except ImportError:
    numpy_installed = False

try:
    import line_profiler
    line_profiler_installed = True
except ImportError:
    line_profiler_installed = False

# is_pypy_interpreter     = True
# is_windows_os           = True
# dateutil_installed      = False
# ultrajson_installed     = False
# line_profiler_installed = False

# determines if excel_com module should be loaded in __init__
loads_excel_module = (not is_pypy_interpreter) and is_windows_os






from .workbook import open_workbook
from .workbook import close_workbook
from .workbook import new_excel_application
from .workbook import any_excel_application
from .workbook import empty_excel_application
from .workbook import all_excel_instances

from .excel_address import col_letter_offset
from .excel_address import col_letter
from .excel_address import col_number

from .classes import excel_levity_cls

__all__ = ['open_workbook',
           'close_workbook',
           'new_excel_application',
           'any_excel_application',
           'empty_excel_application',
           'all_excel_instances',
           'col_letter_offset',
           'col_letter',
           'col_number',
           'excel_levity_cls']




import ctypes
import os
import pythoncom

from _ctypes import COMError as ctypes_error
# noinspection PyUnresolvedReferences
from pythoncom import com_error as pythoncom_error

from ctypes import byref
from ctypes import c_void_p
from ctypes import py_object
from ctypes import POINTER
from ctypes import PyDLL
from ctypes.wintypes import BOOL

from comtypes import IUnknown
from comtypes.client          import CreateObject as comtypes_createobject
from comtypes.automation      import IDispatch    as comtypes_idispatch
from comtypes.client.dynamic  import Dispatch     as comtypes_dispatch

from win32com.client.gencache import EnsureDispatch                          # early-bound references

from ..util.filesystem import standardize_path
from ..util.filesystem import validate_path_exists

from ..util.text import styled
from ..util.text import vengeance_message
from ..util.text import vengeance_warning
from .excel_constants import *

# Windows api functions
FindWindowExA              = ctypes.windll.user32.FindWindowExA
SetForegroundWindow        = ctypes.windll.user32.SetForegroundWindow
AccessibleObjectFromWindow = ctypes.oledll.oleacc.AccessibleObjectFromWindow

# FindWindowExA only accepts ascii strings
xl_main_ascii = 'XLMAIN'.encode('ascii')
xl_desk_ascii = 'XLDESK'.encode('ascii')
excel7_ascii  = 'EXCEL7'.encode('ascii')

corrupt_hwnds = set()
native_om     = -16


def open_workbook(path,
                  excel_app='new',
                  *,
                  read_only=False,
                  update_links=True,
                  windowstate=xlMaximized):

    wb = get_opened_workbook(path)

    if wb is None:
        wb = __open_workbook_dispatch(path,
                                      excel_app,
                                      update_links=update_links,
                                      read_only=read_only,
                                      windowstate=windowstate)

    if (wb.ReadOnly is False) and read_only:
        raise AssertionError("'{}' is not open as read-only".format(wb.Name))

    if wb.ReadOnly and (read_only is False):
        vengeance_warning("'{}' is open as read-only".format(wb.Name))

    return wb


def get_opened_workbook(path):
    """ scan all open workbooks across all Excel instances for matching paths

    :return captured Workbook or None
    """
    path = standardize_path(path, explicit_cwd=False).lower()

    workbooks = []
    for excel_app in all_excel_instances():
        for wb in excel_app.Workbooks:
            wb_path = standardize_path(wb.FullName, explicit_cwd=False).lower()
            wb_name = standardize_path(wb.Name, explicit_cwd=False).lower()

            if path == wb_path:
                workbooks.append(wb)
            elif path == wb_name:
                workbooks.append(wb)

    if workbooks:
        if len(workbooks) > 1:
            vengeance_warning("multiple workbooks matching path: '{}' exist in different "
                              "Excel applications... returning first match".format(path))
        return workbooks[0]

    return None


def is_workbook_open(path):
    wb = get_opened_workbook(path)
    return wb is not None


# noinspection PyUnusedLocal
def close_workbook(wb, save):
    """
    provides a few conveniences that the client may not want to deal with
        1) if Workbook is the only one in the Excel application, closes the application as well
        2) correctly releases the com pointer for the workbook
    """
    import gc

    if save and wb.ReadOnly:
        raise AssertionError("'{}' is open as read-only, cannot save and close".format(wb.Name))

    excel_app = wb.Application
    display_alerts = excel_app.DisplayAlerts

    if save:
        wb.Save()

    excel_app.DisplayAlerts = False
    wb.Close()
    wb = None
    del wb

    if excel_app.Workbooks.Count == 0:
        excel_app.Quit()
        excel_app = None
        del excel_app
    else:
        excel_app.DisplayAlerts = display_alerts

    gc.collect()


def __open_workbook_dispatch(path, excel_app,
                             *,
                             update_links,
                             read_only,
                             windowstate):
    """
    excel_app.DisplayAlerts = False
        what if file not loaded completely?
        what about dialog message?

    excel_app.Workbooks.Open
        other options: password, etc
    """

    path = validate_path_exists(path)
    excel_app = __validate_excel_application(excel_app, windowstate)

    if read_only:
        print(vengeance_message('(opening workbook as read-only ...)'))

    excel_app.DisplayAlerts = False
    wb = excel_app.Workbooks.Open(path, update_links, read_only)
    excel_app.DisplayAlerts = True

    return wb


def __validate_excel_application(excel_app, windowstate=None):
    if excel_app in (None, 'new'):
        return new_excel_application(windowstate)

    if excel_app == 'any':
        return any_excel_application(windowstate)

    if excel_app == 'empty':
        excel_app = empty_excel_application(windowstate)
        if excel_app is None:
            raise AssertionError('empty Excel application not found')

        return excel_app

    if not hasattr(excel_app, 'Workbooks'):
        raise ValueError("excel_app parameter should be in (None, 'new', 'any', 'empty') or an application pointer")

    return excel_app


def new_excel_application(windowstate=None):
    excel_app = comtypes_createobject('Excel.Application', dynamic=True)
    excel_app = __iunknown_pointer_to_python_object(excel_app, comtypes_idispatch)
    excel_app = EnsureDispatch(excel_app)

    reload_all_add_ins(excel_app)
    excel_application_to_foreground(excel_app, windowstate)

    return excel_app


def any_excel_application(windowstate=None):
    excel_app = EnsureDispatch('Excel.Application')
    if excel_app:
        excel_application_to_foreground(excel_app, windowstate)
    else:
        excel_app = new_excel_application(windowstate)

    return excel_app


def empty_excel_application(windowstate=None):
    for excel_app in all_excel_instances():

        if __is_excel_application_empty(excel_app):
            print(vengeance_message('utilizing empty Excel instance: {}'.format(excel_app.Hwsd)))
            excel_application_to_foreground(excel_app, windowstate)

            return excel_app

    return None


def all_excel_instances():
    window_h = __next_window_handle()

    excel_hwnds = set()
    while window_h != 0:
        excel_app = __excel_application_from_window_handle(window_h)

        if excel_app is None:
            window_h = __next_window_handle(window_h)
            continue

        hwnd = excel_app.Hwnd
        if hwnd in excel_hwnds:
            window_h = __next_window_handle(window_h)
            continue

        excel_hwnds.add(hwnd)
        yield excel_app

        window_h = __next_window_handle(window_h)


def reload_all_add_ins(excel_app):
    print(vengeance_message('reloading Excel add-ins...'))

    for add_in in excel_app.AddIns:
        if add_in.Installed:
            name = add_in.Name

            try:
                add_in.Installed = False
                add_in.Installed = True
                print('\t   * {}'.format(name))
            except (ctypes_error, pythoncom_error, NameError) as e:
                vengeance_warning('failed to load Excel add-in: {}\n{}'.format(name, e))

    print()


def excel_application_to_foreground(excel_app, windowstate=None):
    if windowstate not in (None, xlNormal, xlMaximized):
        raise ValueError('windowstate must be in: (None, xlNormal, xlMaximized)')

    if windowstate is not None:
        excel_app.WindowState = windowstate

    excel_app.Visible = True
    SetForegroundWindow(excel_app.Hwnd)


def __next_window_handle(window_h=0):
    return FindWindowExA(0, window_h, xl_main_ascii, None)


# noinspection PyTypeChecker,PyProtectedMember,PyUnresolvedReferences
def __excel_application_from_window_handle(window_h):
    """
    comtypes library is used to search windows handles for Excel application,
    then the application pointer is converted to a pywin object

    obj_ptr = POINTER(comtypes_idispatch)()
        Expected type [_CT], got Type[IDispatch] instead
        this is probably not correct way to set up this pointer

        from pywintypes import IID?
    """
    global corrupt_hwnds

    if window_h in corrupt_hwnds:
        return None

    desk_hwnd   = FindWindowExA(window_h,  None, xl_desk_ascii, None)
    excel7_hwnd = FindWindowExA(desk_hwnd, None, excel7_ascii,  None)

    if excel7_hwnd == 0:
        corrupt_hwnds.add(window_h)
        return None

    obj_ptr = POINTER(comtypes_idispatch)()
    AccessibleObjectFromWindow(excel7_hwnd,
                               native_om,
                               byref(obj_ptr._iid_),
                               byref(obj_ptr))

    try:
        com_ptr   = comtypes_dispatch(obj_ptr).Application
        excel_app = __iunknown_pointer_to_python_object(com_ptr, comtypes_idispatch)
    except (ctypes_error, pythoncom_error, NameError):

        # comtypes Dispatch() call rejected: user may be editing cell
        raise ChildProcessError('Remote Procedure Call to Excel Application rejected. '
                                '\n'
                                '\n\t(Excel will reject automation calls while waiting on user input, '
                                '\n\tcheck if cursor is still active within a cell somewhere)') from None

    try:
        return EnsureDispatch(excel_app)
    except AttributeError:

        # win32com Dispatch() call rejected: COM files may be corrupted
        __move_win32com_gencache_folder()

        raise ChildProcessError('Error dispatching Excel Application from win32com module. '
                                '\n'
                                '\n\t(contents of the win32com gen_py folder may be corrupt)') from None


# noinspection PyTypeChecker
def __iunknown_pointer_to_python_object(com_ptr, interface):
    function_pointer = PyDLL(pythoncom.__file__).PyCom_PyObjectFromIUnknown

    function_pointer.restype  = py_object
    function_pointer.argtypes = (POINTER(IUnknown),
                                 c_void_p,
                                 BOOL)

    # noinspection PyProtectedMember
    return function_pointer(com_ptr._comobj, byref(interface._iid_), True)


# noinspection PyUnusedLocal
def __is_excel_application_empty(excel_app):
    """ an Excel application with no workbooks or a an Excel application
    with the default workbook opened
    """
    for wb in excel_app.Workbooks:
        if wb.Saved:
            return False

        for ws in wb.Sheets:
            if ws.UsedRange.Address != '$A$1':
                return False

    return True


# noinspection PyBroadException
def __move_win32com_gencache_folder():
    """ move win32com gen_py cache files from temp to site-packages folder """
    try:
        import shutil
        import site
        import subprocess
        import win32com
        from pathlib import Path

        old_gcf = os.environ['userprofile'] + '\\AppData\\Local\\Temp\\gen_py\\'
        new_gcf = site.getsitepackages()[1] + '\\win32com\\gen_py\\'

        if not os.path.exists(old_gcf):
            old_gcf = win32com.__gen_path__
            if standardize_path(old_gcf) == standardize_path(new_gcf):
                return

        s = vengeance_message('Attempting to reset win32com gencache folder ...\n')
        s = styled(s, 'yellow', 'bold')
        print(s)

        if os.path.exists(old_gcf):
            subprocess.call('explorer.exe "{}"'.format(Path(old_gcf).parent))
            shutil.rmtree(old_gcf)

        if not os.path.exists(new_gcf):
            subprocess.call('explorer.exe "{}"'.format(Path(new_gcf).parent))
            os.makedirs(new_gcf)

    except Exception:
        pass




from collections import Iterable
from collections import defaultdict

from ..conditional import ordereddict


class IterationDepthError(TypeError):
    pass


def modify_iteration_depth(v, depth=0):
    """
    (this function is heavily utilized to correctly un-nest
     arguments in invariable arity flux_cls methods, so make sure
     you dont fuck anything up if you change anything here !!)

    eg:
        'a'              = modify_iteration_depth([[['a']]], depth=0)
        ['a', 'b']       = modify_iteration_depth([[['a', 'b']]], depth=0)
        [['a']]          = modify_iteration_depth('a', depth=2)
        [[[['a', 'b']]]] = modify_iteration_depth(['a', 'b'], depth=4)

        required behaviors:
            must not reduce depth of any dictionary keys or values
                {'a': None} = modify_iteration_depth([{'a': None}], depth=0)

        undecided behaviors:
            how to deal with mixed iteration depths?
                [['ab', ['cd']]]
                should this return ['ab', ['cd']] or ['ab', 'cd']?

            # has_multiple_depths = set(iteration_depth(_v_, first_element_only=False) for _v_ in v)
    """
    # region {closure functions}
    def is_descendable(value):
        if not is_collection(value):
            return False
        if isinstance(value, dict):
            return False
        if any([is_exhaustable(_v_) for _v_ in value]):
            raise TypeError('cannot modify depth of an exhaustable iterator')

        return (is_subscriptable(value) and
                len(value) == 1)
    # endregion

    nd = iteration_depth(v, first_element_only=True)

    if nd > depth:
        for _ in range(nd - depth):
            if is_descendable(v):
                v = v[0]
            else:
                break
    elif nd < depth:
        for _ in range(depth - nd):
            v = [v]

    return v


def iteration_depth(v, first_element_only=True):
    """
    (this function is heavily utilized to correctly un-nest
     arguments in variable arity flux_cls methods, so make sure
     you dont fuck anything up if you change anything here !!)

    eg:
        0 = iteration_depth('abc')
        1 = iteration_depth(['abc'])
        2 = iteration_depth([[2], [2]])

    eg: items = [1, [2, 2, [3, 3, 3, [4, [5, 5, 5, 5], 4]]]]
        2 = iteration_depth(items, first_element_only=True)
        5 = iteration_depth(items, first_element_only=False)
    """
    if is_exhaustable(v):
        raise TypeError('cannot evaluate iteration depth of an exhaustable value: {}'.format(v))
    if not is_collection(v):
        return 0
    if len(v) == 0:
        return 1

    if isinstance(v, dict):
        v = tuple(v.values())

    if first_element_only:
        return 1 + iteration_depth(v[0], first_element_only)
    else:
        return 1 + max(iteration_depth(_v_, first_element_only) for _v_ in v)


def iterator_to_list(v):
    if is_vengeance_class(v):
        return list(v.rows())

    if is_exhaustable(v):
        return list(v)

    if isinstance(v, range):
        return list(v)

    return v


def is_collection(v):
    """ determine if value is an iterable object or data structure
    function used mainly to distinguish data structures from other iterables

    eg:
        False = is_collection('mike')
        True  = is_collection(['m', 'i' 'k' 'e'])
    """
    if isinstance(v, (str, bytes, range)):
        return False

    return isinstance(v, Iterable)


# noinspection PyStatementEffect,PyBroadException
def is_subscriptable(o):
    try:
        o[0]
        return True
    except Exception:
        return False


def is_exhaustable(o):
    return hasattr(o, '__next__')


def standardize_variable_arity_arguments(v, depth):
    # region {closure function}
    def exhaust_iterator(a):
        if is_subscriptable(a) and (is_exhaustable(a[0]) or isinstance(a[0], range)):
            return iterator_to_list(a[0])

        return iterator_to_list(a)
    # endregion

    v = exhaust_iterator(v)
    v = modify_iteration_depth(v, depth)
    v = exhaust_iterator(v)

    return v


def is_vengeance_class(o):
    bases = set(base_class_names(o))
    return bool(bases & {'flux_cls',
                         'excel_levity_cls'})


def is_namedtuple_class(o):
    return isinstance(o, tuple) and (type(o) is not tuple)


def base_class_names(o):
    return [b.__name__ for b in o.__class__.mro()]


def transpose(m, astype=tuple):
    if astype is tuple:
        return transpose_as_tuples(m)
    if astype is list:
        return transpose_as_lists(m)

    raise TypeError('astype parameter must be either tuple or list')


def transpose_as_tuples(m):
    if iteration_depth(m) == 1:
        return ((v,) for v in m)
    else:
        return zip(*m)


def transpose_as_lists(m):
    if iteration_depth(m) == 1:
        return ([v] for v in m)
    else:
        return (list(v) for v in zip(*m))


def map_to_numeric_indices(sequence, start=0):
    """ :return {value: index} for all items in sequence

    values are modified before they are added as keys:
        non-unique keys are coerced to string and appended with '_{num}'
        suffix to ensure that indices are incremented correctly

    eg
        {'a':   0,
         'b':   1,
         'c':   2,
         'b_2': 3} = map_numeric_indices(['a', 'b', 'c', 'b'])
    """
    indices   = ordereddict()
    nonunique = defaultdict(lambda: 1)

    for i, v in enumerate(sequence, start):
        if isinstance(v, bytes):
            v_s = v.decode('utf-8')
        else:
            v_s = str(v)

        if v_s in indices:
            nonunique[v_s] += 1
            v_s = '{}_{}'.format(v_s, nonunique[v_s])

        indices[v_s] = i

    return indices


class OrderedDefaultDict(ordereddict):

    def __init__(self, default):
        self.default_factory = default
        super().__init__()

    def append_items(self, items):
        """ all items that map to the same key are appended to a list """
        if self.default_factory is not list:
            raise TypeError('self.default_factory function must be list')

        if isinstance(items, dict):
            items = items.items()

        for item in items:
            try:
                k, v = item
            except Exception:
                raise IndexError('items must be (key, value) pairs') from None

            self[k].append(v)

        return self

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)

        self[key] = value = self.default_factory()
        return value

    def defaultdict(self):
        return defaultdict(self.default_factory, self.items())

    def ordereddict(self):
        return ordereddict(self.items())




import os
import sys

from textwrap import dedent
from logging import Logger
from logging import Formatter
from logging import FileHandler
from logging import StreamHandler
from logging import (NOTSET,
                     DEBUG,
                     INFO,
                     WARNING,
                     ERROR,
                     CRITICAL)

from .. util.filesystem import parse_path
from .. util.filesystem import standardize_dir
from .. util.text import object_name
from .. util.text import styled

from .. conditional import is_tty_console


class log_cls(Logger):
    banner_character = '*'
    banner_width     = None

    def __init__(self, path_or_name='',
                       level='DEBUG',
                       log_format='%(message)s',
                       exception_callback=None,
                       colored_statements=True):
        """
        :param path_or_name:
            parsed to determine name of logger
            if path_or_name includes directory or file extension, a file handler is added
        :param log_format:
            format to be applied to handlers
            eg log_format:
                '[%(asctime)s] [%(levelname)s] %(message)s'
        :param exception_callback:
            function to be invoked when self.exception_handler() is called
            sys.excepthook is automatically set to self.exception_handler if
            exception_callback is a valid function
        """
        if (exception_callback is not None) and not callable(exception_callback):
            raise TypeError('exception_callback must be callable')

        p_path = parse_path(path_or_name, explicit_cwd=False)

        if isinstance(level, str):
            level = level.upper().strip()

        super().__init__(p_path.filename, level)

        self.path      = self.__set_path(p_path.directory, p_path.filename, p_path.extension)
        self.formatter = Formatter(log_format)

        self.exception_callback = exception_callback
        self.exception_message  = ''

        self._add_stream_handler(sys.stdout, colored_statements)
        self._add_file_handler(self.path)

        if exception_callback:
            sys.excepthook = self.exception_handler

    @property
    def stream_handlers(self):
        handlers = []
        for h in self.handlers:
            if type(h) in (StreamHandler, colored_streamhandler_cls):
                handlers.append(h)

        return handlers

    @property
    def file_handlers(self):
        handlers = []
        for h in self.handlers:
            if isinstance(h, FileHandler):
                handlers.append(h)

        return handlers

    def set_format(self, log_format):
        self.formatter = Formatter(log_format)
        for h in self.handlers:
            h.setFormatter(self.formatter)

    def add_parent_log(self, p_log):
        if id(p_log) == id(self):
            raise ValueError('parent log and self are the same')

        self.parent = p_log
        self.close_stream_handlers()

    def close(self):
        self.close_stream_handlers()
        self.close_file_handlers()

    def close_stream_handlers(self):
        for h in self.stream_handlers:
            h.close()
            self.removeHandler(h)

    def close_file_handlers(self):
        for h in self.file_handlers:
            h.close()
            self.removeHandler(h)

    def _add_stream_handler(self, stream, colored_statements):
        if not colored_statements:
            h = StreamHandler(stream)
        elif is_tty_console:                    # TTY console doesn't support ascii escapes
            h = StreamHandler(stream)
        else:
            h = colored_streamhandler_cls(stream)

        h.setLevel(self.level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def _add_file_handler(self, path):
        if not path:
            return

        h = FileHandler(path, mode='w', encoding='utf-8')

        h.setLevel(self.level)
        h.setFormatter(self.formatter)
        self.addHandler(h)

    def exception_handler(self, e_type, e_msg, e_traceback):
        """
        sys.excepthook = log.exception_handler

        try:
            1 / 0
        except:
            log.exception_handler(*sys.exc_info())
        """

        self.exception_message = self.__formatted_exception_message(e_type, e_msg, e_traceback)
        self.error(self.exception_message, exc_info=(e_type, e_msg, e_traceback))

        if self.exception_callback:
            self.exception_callback()

        return self.exception_message

    def __formatted_exception_message(self, e_type, e_msg, e_traceback):
        title_message = '(The result w:resign was added to the game information ...)'

        _e_type_ = 'Exception'
        _e_msg_  = 'unknown error'
        filename = 'unknown'
        lineno   = 'unknown'

        if e_type:
            _e_type_ = object_name(e_type)

        if e_msg:
            _e_msg_ = '{}'.format(str(e_msg).replace('"', "'"))

        if e_traceback:
            filename = e_traceback.tb_frame.f_code.co_filename
            lineno   = e_traceback.tb_lineno

        if isinstance(self.banner_width, int):
            banner_width = self.banner_width
        else:
            banner_width = max([len(title_message),
                              *[len(line) for line in str(_e_msg_).split('\n')],
                              *[len(line) for line in str(filename).split('\n')]]) + 10
            banner_width = max(banner_width, 90)

        banner = self.banner_character * banner_width

        exception_message = dedent('''
        {banner}
            {title_message}
            {repr_self}
            
            <{e_type}> {e_msg}
            File: {filename}
            Line: {lineno}
        {banner}
        
        ''').format(banner=banner,
                    title_message=title_message,
                    repr_self=repr(self),
                    e_type=_e_type_,
                    e_msg=_e_msg_,
                    filename=filename,
                    lineno=lineno)

        return exception_message

    @staticmethod
    def __set_path(filedir, filename, extension):
        if not (filedir or extension):
            return ''

        filedir = standardize_dir(filedir, explicit_cwd=True)
        if not os.path.exists(filedir):
            os.makedirs(filedir)

        if extension == '.py':
            extension = '.log'

        path = filedir + filename + (extension or '.log')

        return path

    def __repr__(self):
        return 'vengeance log: {}'.format(self.name)


class colored_streamhandler_cls(StreamHandler):
    level_colors  = {NOTSET:   'grey',
                     DEBUG:    'grey',
                     INFO:     'white',
                     WARNING:  'bright yellow',
                     ERROR:    'red',
                     CRITICAL: 'bright magenta'}

    def __init__(self, stream=None):
        super().__init__(stream)

    def emit(self, record):
        # noinspection PyBroadException
        try:
            s = self.format(record) + self.terminator
            s = styled(s, self.level_colors.get(record.levelno, 'grey'), 'bold')

            self.stream.write(s)
            self.flush()

        except RecursionError:
            raise
        except Exception:
            self.handleError(record)

